@@ -4,6 +4,7 @@ from adbase import *
 
 class ADObject(ADBase):
     """Python object that represents any active directory object."""
+
     _ldap_adsi_obj = None
     _gc_adsi_obj = None
     _schema_adsi_obj = None
@@ -11,43 +12,109 @@ class ADObject(ADBase):
     _mandatory_attributes = None
     _optional_attributes = None
     _py_ad_object_mappings = {}
-
+    
+    def __set_adsi_obj(self, distinguished_name):
+        """Internal method that creates the connection to the backend ADSI object."""
+    
+        if self.default_username and self.default_password:
+            # from http://msdn.microsoft.com/en-us/library/windows/desktop/aa706065(v=vs.85).aspx
+            # With the LDAP provider for Active Directory, you may pass in
+            # lpszUserName as one of the following strings:
+            # (1) The name of a user account, such as "jeffsmith". To use a user name 
+            # by itself, you must set only the ADS_SECURE_AUTHENTICATION flag 
+            # in the lnReserved parameter.
+            # (2) The user path from a previous version of Windows NT, such 
+            # as "Fabrikam\jeffsmith".
+            # (3) Distinguished Name, such as "CN=Jeff Smith, OU=Sales,
+            # DC=Fabrikam,DC=Com". To use a DN, the lnReserved parameter
+            # must be zero or it must include the ADS_USE_SSL flag
+            # (4) User Principal Name (UPN), such as "jeffsmith@Fabrikam.com".
+            # To use a UPN, you must assign the appropriate UPN value for the
+            # userPrincipalName attribute of the target user object.
+            
+            # In order to be consistent (and because troubleshooting this
+            # is horrid), we're just going to force user name to be of form
+            # (1) or (4) and document it. Offhand, I'm not seeing any
+            # use cases where (3) would allow something not possible in
+            # the combinations of options (1), or (4).
+            _ds = self.adsi_provider.getObject('', "LDAP:")
+            if self.default_ldap_authentication_flag > 0:
+                flag = self.default_ldap_authentication_flag
+            else:
+                # I'm choosing to force encryption of the login credentials.
+                # This does not require SSL to be configured, so I believe this
+                # should work for everyone. If not, we can change later. 
+                flag = ADS_AUTHENTICATION_TYPE['ADS_SECURE_AUTHENTICATION']
+                if self.default_ssl:
+                    flag = flag | ADS_AUTHENTICATION_TYPE['ADS_USE_ENCRYPTION']
+            self._ldap_adsi_obj = _ds.OpenDSObject(
+                    self.__ads_path,
+                    self.default_username,
+                    self.default_password,
+                    flag)
+            
+        elif self.default_ssl:
+            raise Exception("Using SSL without specifying credentials is currently unsupported due to what appears to be a bug in pywin32.")
+            # from: http://msdn.microsoft.com/en-us/library/windows/desktop/aa772247(v=vs.85).aspx
+            # If ADS_USE_SSL is not combined with the ADS_SECURE_AUTHENTICATION
+            # flag and the supplied credentials are NULL, the bind will be
+            # performed anonymously. If ADS_USE_SSL is combined with the 
+            # ADS_SECURE_AUTHENTICATION flag and the supplied credentials 
+            # are NULL, then the credentials of the calling thread are used.
+            flag = ADS_AUTHENTICATION_TYPE['ADS_SECURE_AUTHENTICATION'] | \
+                            ADS_AUTHENTICATION_TYPE['ADS_USE_ENCRYPTION']
+            _ds = self.adsi_provider.getObject('', "LDAP:")
+            self._ldap_adsi_obj = _ds.OpenDSObject(
+                            self.__ads_path,
+                            None, # username
+                            None, # password
+                            flag)
+        else:
+            self._ldap_adsi_obj = self.adsi_provider.getObject('', self.__ads_path)
+    
     def __init__(self, distinguished_name=None, adsi_ldap_com_object=None, options={}):
         if adsi_ldap_com_object:
             self._ldap_adsi_obj = adsi_ldap_com_object
         elif distinguished_name:
-            if 'server' in options:
-                self.default_ldap_server = options['server']
-            if 'port' in options:
-                self.default_ldap_port = options['port']
-            self.__ads_path = pyadutils.generate_ads_path(distinguished_name, 'LDAP', self.default_ldap_server, self.default_ldap_port)
-            try:
-                self._ldap_adsi_obj = self.adsi_provider.getObject('',self.__ads_path)
-            except pywintypes.com_error, excpt:
-                additional_info = {
-                    'distinguished_name':distinguished_name,
-                    'server':self.default_ldap_server,
-                    'port':self.default_ldap_port
-                }
-                pyadutils.pass_up_com_exception(excpt, additional_info)
+            self._set_defaults(options)
+            self.__ads_path = pyadutils.generate_ads_path(distinguished_name,
+                            self.default_ldap_protocol,
+                            self.default_ldap_server,
+                            self.default_ldap_port
+            )
+            self.__set_adsi_obj(distinguished_name)
         else:
             raise Exception("Either a distinguished name or a COM object must be provided to create an ADObject")
 
-        # by pulling the DN from object instead of what is passed in, we gaurantee correct capitalization
+        # by pulling the DN from object instead of what is passed in,
+        # we guarantee correct capitalization
         self.__distinguished_name = self.get_attribute('distinguishedName', False)
-        self.__object_guid = pyadutils.convert_guid(self.get_attribute('objectGUID', False))
-        # Set pyAD Object Type
+        self.__object_guid = self.get_attribute('objectGUID', False)
+        if self.__object_guid is not None:
+            self.__object_guid = pyadutils.convert_guid(self.__object_guid)
+        # Set pyAD Object Type        
         occn = self.get_attribute('objectCategory',False)
         if occn:
             # pull out CN from DN
             object_category_cn = occn.split('=',1)[1].split(",",1)[0]
-            # some object categories are not very human readable so we provide the option to override
+            # some object categories are not very human readable
+            # so we provide the option to override
             if object_category_cn in PYAD_CATEGORY_TYPE_OVERRIDE_MAPPPINGS:
                 self._type = PYAD_CATEGORY_TYPE_OVERRIDE_MAPPPINGS[object_category_cn]
             else:
                 self._type = object_category_cn.lower()
         else:
-            self._type = 'unknown'
+            # Sometimes you don't have access to objectCategory attribute,
+            # try, with objectClass attribute
+            objClass = self.get_attribute('objectClass',True)
+            if 'domain' in objClass:
+                self._type = 'domain'
+            elif 'user' in objClass:
+                self._type = 'user'
+            elif 'organizationalUnit' in objClass:
+                self._type = 'organizationalUnit'
+            else:
+                self._type = 'unknown'
 
     @classmethod
     def from_guid(cls, guid, options={}):
@@ -77,14 +144,22 @@ class ADObject(ADBase):
         sid = self.objectSid
         return pyadutils.convert_sid(sid) if sid else None
 
-    dn = property(fget=lambda self: self.__distinguished_name, doc="Distinguished Name (DN) of the object")
-    prefixed_cn = property(fget=__get_prefixed_cn, doc="Prefixed CN (such as 'cn=mycomputer' or 'ou=mycontainer' of the object")
-    guid = property(fget=lambda self: self.__object_guid, doc="Object GUID of the object")
-    adsPath = property(fget=lambda self: self.__ads_path, doc="ADsPath of Active Directory object (such as 'LDAP://cn=me,...,dc=com'")
-    type = property(fget=lambda self: self._type, doc="pyAD object type (user, computer, group, organizationalUnit, domain).")
-    parent_container_path = property(fget=lambda self: self.dn.split(',',1)[1], doc="Returns the DN of the object's parent container.")
-    guid_str = property(fget=lambda self: str(self.guid)[1:-1], doc="Object GUID of the object")
-    sid = property(fget=__get_object_sid, doc='Get the SID of the Active Directory object')
+    dn = property(fget=lambda self: self.__distinguished_name,
+                    doc="Distinguished Name (DN) of the object")
+    prefixed_cn = property(fget=__get_prefixed_cn,
+                    doc="Prefixed CN (such as 'cn=mycomputer' or 'ou=mycontainer' of the object")
+    guid = property(fget=lambda self: self.__object_guid,
+                    doc="Object GUID of the object")
+    adsPath = property(fget=lambda self: self.__ads_path,
+                    doc="ADsPath of Active Directory object (such as 'LDAP://cn=me,...,dc=com'")
+    type = property(fget=lambda self: self._type,
+                    doc="pyAD object type (user, computer, group, organizationalUnit, domain).")
+    parent_container_path = property(fget=lambda self: self.dn.split(',',1)[1],
+                    doc="Returns the DN of the object's parent container.")
+    guid_str = property(fget=lambda self: str(self.guid)[1:-1],
+                    doc="Object GUID of the object")
+    sid = property(fget=__get_object_sid,
+                    doc='Get the SID of the Active Directory object')
 
     def __hash__(self):
         # guid is always unique so that we can depend on that for providing a unique hash
@@ -96,8 +171,10 @@ class ADObject(ADBase):
     __repr__ = __str__
 
     def __cmp__(self, other):
-        # it doesn't make sense why you'd ever have to decide if one GUID was larger than the other,
-        # but it's important to be able to know if two pyAD objects represent the same AD object.
+        # it doesn't make sense why you'd ever have to decide
+        # if one GUID was larger than the other,
+        # but it's important to be able to know if two
+        # pyAD objects represent the same AD object.
         if (self.guid == other.guid):
             return 0
         elif (self.guid < other.guid):
@@ -116,12 +193,56 @@ class ADObject(ADBase):
     def _flush(self):
         "Commits any changes to the AD object."
         return self._ldap_adsi_obj.SetInfo()
-
-    def _init_global_catalog_object(self, options={}):
+        
+    def __set_gc_adsi_obj(self):
+        if self._gc_adsi_obj:
+            return
+        if self.default_ldap_authentication_flag > 0:
+            _ds = self.adsi_provider.getObject('', self.default_ldap_protocol+":")
+            path = pyadutils.generate_ads_path(
+                        self.dn,
+                        'GC',
+                        self.default_gc_server,
+                        self.default_gc_port
+            )
+            self._ldap_adsi_obj = _ds.OpenDSObject(path,
+                            self.default_username,
+                            self.default_password,
+                            self.default_ldap_authentication_flag
+            )
+        else:
+            self._gc_adsi_obj = self.adsi_provider.GetObject('',
+                    pyadutils.generate_ads_path(self.dn, 'GC', 
+                            options.get('server'), options.get('port')
+                    )
+            )
+            
+    def __set_gc_adsi_obj(self):
+        path = pyadutils.generate_ads_path(
+                        self.dn,
+                        'GC',
+                        self.default_gc_server,
+                        self.default_gc_port
+        )
+        if self.default_username and self.default_password:
+            _ds = self.adsi_provider.getObject('', "LDAP:")
+            flag = ADS_AUTHENTICATION_TYPE['ADS_SECURE_AUTHENTICATION']
+            if self.default_ssl:
+                flag = flag | ADS_AUTHENTICATION_TYPE['ADS_USE_ENCRYPTION']
+            self._gc_adsi_obj = _ds.OpenDSObject(
+                    path,
+                    self.default_username,
+                    self.default_password,
+                    flag)
+        else:
+            self._gc_adsi_obj = self.adsi_provider.GetObject('', path)
+    
+    def _init_global_catalog_object(self, force=False, options={}):
         """Initializes the global catalog ADSI com object to be
         used when querying the global catalog instead of the domain directly."""
-        if not self._gc_adsi_obj:
-            self._gc_adsi_obj = self.adsi_provider.GetObject('',pyadutils.generate_ads_path(self.dn, 'GC', options.get('server'), options.get('port')))
+        if not self._gc_adsi_obj and not force:
+            self._set_defaults(options)
+            self.__set_gc_adsi_obj()
 
     def _init_schema_object(self):
         if not self._schema_adsi_obj:
@@ -221,14 +342,16 @@ class ADObject(ADBase):
     def append_to_attribute(self, attribute, valuesToAppend):
         """Appends values in list valuesToAppend to the specified multi-valued attribute.
         valuesToAppend can contain a single value or a list of multiple values."""
-        difference = list(set(pyadutils.generate_list(valuesToAppend)) - set(self.get_attribute(attribute)))
+        difference = list(set(pyadutils.generate_list(valuesToAppend)) \
+                        - set(self.get_attribute(attribute)))
         if len(difference) != 0:
             self._set_attribute(attribute,3,difference)
             self._flush()
 
     def remove_from_attribute(self, attribute, valuesToRemove):
         """Removes any values in list valuesToRemove from the specified multi-valued attribute."""
-        difference = list(set(pyadutils.generate_list(valuesToRemove)) & set(self.get_attribute(attribute)))
+        difference = list(set(pyadutils.generate_list(valuesToRemove)) \
+                        & set(self.get_attribute(attribute)))
         if len(difference) != 0:
             self._set_attribute(attribute,4,difference)
             self._flush()
@@ -260,7 +383,7 @@ class ADObject(ADBase):
                 nv = self.get_attribute('userAccountControl',False) ^ ADS_USER_FLAG[userFlag]
             else:
                 nv = self.get_attribute('userAccountControl',False)
-            # i f the flag is true, then the value is present and
+            # if the flag is true, then the value is present and
             # we add it to the starting point with B-OR.
             # Otherwise, if it's false, it's just not present,
             # so we leave it without any mention of the flag as in previous step.
@@ -269,6 +392,7 @@ class ADObject(ADBase):
             self.update_attribute('userAccountControl',nv)
 
     def disable(self):
+        """Disables the user or computer"""
         try:
             if self._ldap_adsi_obj.AccountDisabled == False:
                 self._ldap_adsi_obj.AccountDisabled = True
@@ -277,6 +401,7 @@ class ADObject(ADBase):
             pyadutils.pass_up_com_exception(excpt)
 
     def enable(self):
+        """Enables the user or computer"""
         try:
             if self._ldap_adsi_obj.AccountDisabled == True:
                 self._ldap_adsi_obj.AccountDisabled = False
@@ -294,15 +419,17 @@ class ADObject(ADBase):
 
         new_ou_object expects a ADContainer object where the current object will be moved to."""
         try:
-            new_ou_object._ldap_adsi_obj.MoveHere(('LDAP://'+self.dn), self.prefixed_cn)
+            new_path = self.default_ldap_protocol + '://' + self.dn
+            new_ou_object._ldap_adsi_obj.MoveHere(new_path, self.prefixed_cn)
             new_ou_object._flush()
         except pywintypes.com_error, excpt:
             pyadutils.pass_up_com_exception(excpt)
         new_dn = ','.join((self.prefixed_cn, new_ou_object.dn))
         time.sleep(.5)
-        self.__ads_path = pyadutils.generate_ads_path(new_dn, 'LDAP',
+        self.__ads_path = pyadutils.generate_ads_path(new_dn, self.default_ldap_protocol,
                 self.default_ldap_server, self.default_ldap_port)
-        self._ldap_adsi_obj = self.adsi_provider.getObject('', self.__ads_path)
+        self.__set_adsi_obj()
+        self.__set_gc_adsi_obj()
         self.__distinguished_name = self.get_attribute('distinguishedName', False)
 
     def rename(self, new_name, set_sAMAccountName=True):
@@ -317,18 +444,19 @@ class ADObject(ADBase):
         try:
             if self.type in ('user', 'computer', 'group') and set_sAMAccountName:
                 self._ldap_adsi_obj.Put('sAMAccountName', new_name)
-            parent._ldap_adsi_obj.MoveHere(('LDAP://' + self.dn), pcn)
+            new_path = self.default_ldap_protocol+'://' + self.dn
+            parent._ldap_adsi_obj.MoveHere(new_path, pcn)
             parent._flush()
         except pywintypes.com_error, excpt:
             pyadutils.pass_up_com_exception(excpt)
         new_dn = ','.join((pcn, parent.dn))
         time.sleep(.5)
-        self.__ads_path = pyadutils.generate_ads_path(new_dn, 'LDAP',
+        self.__ads_path = pyadutils.generate_ads_path(new_dn, self.default_ldap_protocol,
                 self.default_ldap_server, self.default_ldap_port)
-        self._ldap_adsi_obj = self.adsi_provider.getObject('', self.__ads_path)
+        self.__set_adsi_obj()
+        self.__set_gc_adsi_obj()
         self.__distinguishedName = self.get_attribute('distinguishedName', False)
 
-
     def add_to_group(self, group):
         """Adds current object to the specified group.
         group expects an ADGroup object."""
@@ -352,6 +480,7 @@ class ADObject(ADBase):
         return self.set_manager(None)
 
     def dump_to_xml(self, whitelist_attributes=[], blacklist_attributes=[]):
+        """Dumps object and all human-readable attributes to an xml document which is returned as a string."""
         if len(whitelist_attributes) == 0:
             whitelist_attributes = self.get_allowed_attributes()
         attributes = list(set(whitelist_attributes) - set(blacklist_attributes))
@@ -396,7 +525,6 @@ class ADObject(ADBase):
         return doc.toxml(encoding="UTF-8")
 
     def adjust_pyad_type(self):
-        """Adjusts pyAD class to match object pyad type."""
         if self.type in self._py_ad_object_mappings.keys():
             self.__class__ = self._py_ad_object_mappings[self.type]
         else:
@@ -404,12 +532,13 @@ class ADObject(ADBase):
 
     def __get_parent_container(self):
         q = ADObject.from_dn(self.parent_container_path,
-            options={'server':self.default_ldap_server,'port':self.default_ldap_port})
+                options = self._make_options())
         q.adjust_pyad_type()
         return q
-    parent_container = property(__get_parent_container)
+    parent_container = property(__get_parent_container, doc="Object representing the container in which this object lives")
 
     def delete(self):
+        """Deletes the object from the domain"""
         parent = self.parent_container
         if not parent:
             raise Exception("Object does not have a parent container. Cannot be deleted")
